@@ -7,11 +7,8 @@
 #include "types.h"
 #include "links.h"
 #include "internal.h"
-#include "checked-integer-arithmetic.h"
 
-zathura_index_element_t*
-zathura_index_element_new(const char* title)
-{
+zathura_index_element_t* zathura_index_element_new(const char* title) {
   if (title == NULL) {
     return NULL;
   }
@@ -26,9 +23,7 @@ zathura_index_element_new(const char* title)
   return res;
 }
 
-void
-zathura_index_element_free(zathura_index_element_t* index)
-{
+void zathura_index_element_free(zathura_index_element_t* index) {
   if (index == NULL) {
     return;
   }
@@ -38,9 +33,7 @@ zathura_index_element_free(zathura_index_element_t* index)
   g_free(index);
 }
 
-zathura_image_buffer_t*
-zathura_image_buffer_create(unsigned int width, unsigned int height)
-{
+zathura_image_buffer_t* zathura_image_buffer_create(unsigned int width, unsigned int height) {
   g_return_val_if_fail(width != 0, NULL);
   g_return_val_if_fail(height != 0, NULL);
 
@@ -49,15 +42,15 @@ zathura_image_buffer_create(unsigned int width, unsigned int height)
     return NULL;
   }
 
-  zathura_image_buffer_t* image_buffer = malloc(sizeof(zathura_image_buffer_t));
+  zathura_image_buffer_t* image_buffer = g_try_malloc(sizeof(zathura_image_buffer_t));
   if (image_buffer == NULL) {
     return NULL;
   }
 
-  image_buffer->data = calloc(size, sizeof(unsigned char));
+  image_buffer->data = g_try_malloc0_n(size, sizeof(unsigned char));
 
   if (image_buffer->data == NULL) {
-    free(image_buffer);
+    g_free(image_buffer);
     return NULL;
   }
 
@@ -68,33 +61,26 @@ zathura_image_buffer_create(unsigned int width, unsigned int height)
   return image_buffer;
 }
 
-void
-zathura_image_buffer_free(zathura_image_buffer_t* image_buffer)
-{
+void zathura_image_buffer_free(zathura_image_buffer_t* image_buffer) {
   if (image_buffer == NULL) {
     return;
   }
 
-  free(image_buffer->data);
-  free(image_buffer);
+  g_free(image_buffer->data);
+  g_free(image_buffer);
 }
 
-static void
-document_information_entry_free(void* data)
-{
+static void document_information_entry_free(void* data) {
   zathura_document_information_entry_t* entry = data;
   zathura_document_information_entry_free(entry);
 }
 
-girara_list_t*
-zathura_document_information_entry_list_new(void)
-{
-  return girara_list_new2(document_information_entry_free);
+girara_list_t* zathura_document_information_entry_list_new(void) {
+  return girara_list_new_with_free(document_information_entry_free);
 }
 
-zathura_document_information_entry_t*
-zathura_document_information_entry_new(zathura_document_information_type_t type, const char* value)
-{
+zathura_document_information_entry_t* zathura_document_information_entry_new(zathura_document_information_type_t type,
+                                                                             const char* value) {
   if (value == NULL) {
     return NULL;
   }
@@ -110,9 +96,7 @@ zathura_document_information_entry_new(zathura_document_information_type_t type,
   return entry;
 }
 
-void
-zathura_document_information_entry_free(zathura_document_information_entry_t* entry)
-{
+void zathura_document_information_entry_free(zathura_document_information_entry_t* entry) {
   if (entry == NULL) {
     return;
   }
